# -*- coding: utf-8 -*-

import hashlib
import html5lib
import re
from collections import Counter, defaultdict
from lxml import etree
from lxml.html import tostring
from lxml.cssselect import CSSSelector

from .. import config
from ..DefaultOrderedDict import DefaultOrderedDict
from ..messages import *

def unescape(string):
    import html
    return html.unescape(string)


def findAll(sel, context):
    if isinstance(context, constants.specClass):
        context = context.document
    try:
        return CSSSelector(sel, namespaces={"svg":"http://www.w3.org/2000/svg"})(context)
    except Exception as e:
        die("The selector '{0}' returned an error:\n{1}", sel, e)
        return []


def find(sel, context=None):
    result = findAll(sel, context)
    if result:
        return result[0]
    else:
        return None


def escapeCSSIdent(val):
    if len(val) == 0:
        die("Programming error: can't escape an empty ident.")
        return ""
    ident = ""
    firstCode = val[0]
    for i,code in enumerate(ord(x) for x in val):
        if code == 0:
            die("Invalid character: the string '{0}' somehow has a NUL in it.", val)
            return ""
        if (0x1 <= code <= 0x1f or
            code == 0x7f or
            (i == 0 and 0x30 <= code <= 0x39) or
                (i == 1 and 0x30 <= code <= 0x39 and firstCode == 0x2d)):
            ident += r"\{0:x} ".format(code)
        elif (code >= 0x80 or
              code == 0x2d or
              code == 0x5f or
              0x30 <= code <= 0x39 or
              0x41 <= code <= 0x5a or
              0x61 <= code <= 0x7a):
            ident += chr(code)
        else:
            ident += r"\{0}".format(chr(code))
    return ident


def escapeUrlFrag(val):
    result = ""
    for char in val:
        if validUrlUnit(char):
            result += char
        else:
            for b in char.encode("utf-8"):
                result += "%{:0>2x}".format(b)
    return result


def validUrlUnit(char):
    c = ord(char)
    if c < 0xa0:
        # ASCII range
        if (
            c == 0x21 or
            c == 0x24 or
            0x26 <= c <= 0x29 or
            0x2a <= c <= 0x3b or
            c == 0x3d or
            0x3f <= c <= 0x5a or
            c == 0x5f or
            0x61 <= c <= 0x7a or
            c == 0x7e):
            return True
        return False
    else:
        if (0xd800 <= c <= 0xdfff or
            0xfdd0 <= c <= 0xfdef):
            return False
        if (c % 0xffff) in [0xfffe, 0xffff]:
            # Last two bytes are FFFE or FFFF
            return False
        return True


def textContent(el, exact=False):
    # If exact is False, then any elements with data-deco attribute
    # get ignored in the textContent.
    # This allows me to ignore things added by Bikeshed by default.
    if len(el) == 0:
        return el.text or ''
    if exact:
        return tostring(el, method='text', with_tail=False, encoding="unicode")
    else:
        return textContentIgnoringDecorative(el)


def textContentIgnoringDecorative(el):
    str = el.text or ''
    for child in childElements(el):
        if child.get("data-deco") is None:
            str += textContentIgnoringDecorative(child)
        str += child.tail or ''
    return str


def innerHTML(el):
    if el is None:
        return ''
    return (el.text or '') + ''.join(tostring(x, encoding="unicode") for x in el)


def outerHTML(el, literal=False):
    if el is None:
        return ''
    if isinstance(el, str):
        return el
    if isinstance(el, list):
        return "".join(map(outerHTML, el))
    if el.get("bs-autolink-syntax") is not None and not literal:
        return el.get("bs-autolink-syntax")
    return tostring(el, with_tail=False, encoding="unicode")

def serializeTag(el):
    # Serialize *just* the opening tag for the element.
    # Use when you want to output the HTML,
    # but it might be a container with a lot of content.
    tag = "<" + el.tag
    for n,v in el.attrib.items():
        tag += ' {n}="{v}"'.format(n=n, v=escapeAttr(v))
    tag += ">"
    return tag


def foldWhitespace(text):
    return re.sub(r"(\s|\xa0)+", " ", text)


def parseHTML(text):
    doc = html5lib.parse(text, treebuilder='lxml', namespaceHTMLElements=False)
    head = doc.getroot()[0]
    body = doc.getroot()[1]
    if len(body) or body.text is not None:
        # Body contains something, so return that.
        contents = [body.text] if body.text is not None else []
        contents.extend(body.iterchildren())
        return contents
    elif len(head) or head.text is not None:
        # Okay, anything in the head?
        contents = [head.text] if head.text is not None else []
        contents.extend(head.iterchildren())
        return contents
    else:
        return []


def parseDocument(text):
    doc = html5lib.parse(text, treebuilder='lxml', namespaceHTMLElements=False)
    return doc


def escapeHTML(text):
    # Escape HTML
    return text.replace('&', '&amp;').replace('<', '&lt;')


def escapeAttr(text):
    return text.replace('&', '&amp;').replace("'", '&apos;').replace('"', '&quot;')


def clearContents(el):
    del el[:]
    el.text = ''
    return el


def parentElement(el):
    return el.getparent()


def appendChild(parent, *children):
    # Appends either text or an element.
    children = list(config.flatten(children))
    for child in children:
        if isinstance(child, str):
            if len(parent) > 0:
                parent[-1].tail = (parent[-1].tail or '') + child
            else:
                parent.text = (parent.text or '') + child
        else:
            if len(parent) == 0 and parent.text is not None:
                # LXML "helpfully" assumes you meant to insert it before the text,
                # and so moves the text into the element's tail when you append.
                text, parent.text = parent.text, None
                parent.append(child)
                parent.text = text
            else:
                # For some reason it doesn't make any weird assumptions about text
                # when the parent already has children; the last child's tail
                # doesn't get moved into the appended child or anything.
                parent.append(child)
    return children[-1] if len(children) else None


def prependChild(parent, child):
    # Prepends either text or an element to the parent.
    if isinstance(child, str):
        if parent.text is None:
            parent.text = child
        else:
            parent.text = child + parent.text
    else:
        removeNode(child)
        parent.insert(0, child)
        if parent.text is not None:
            child.tail = (child.tail or '') + parent.text
            parent.text = None


def insertBefore(target, *els):
    parent = target.getparent()
    index = parent.index(target)
    prevSibling = parent[index - 1] if index > 0 else None
    for el in els:
        if isinstance(el, str):
            if prevSibling is not None:
                prevSibling.tail = (prevSibling.tail or '') + el
            else:
                parent.text = (parent.text or '') + el
        else:
            parent.insert(index, el)
            index += 1
            prevSibling = el
    return target


def insertAfter(target, *els):
    parent = target.getparent()
    for el in els:
        if isinstance(el, str):
            target.tail = (target.tail or '') + el
        else:
            parent.insert(parent.index(target) + 1, el)
            target = el
    return target


def removeNode(node):
    parent = node.getparent()
    if parent is None:
        return node
    index = parent.index(node)
    text = node.tail or ''
    node.tail = None
    if index == 0:
        parent.text = (parent.text or '') + text
    else:
        prevsibling = parent[index - 1]
        prevsibling.tail = (prevsibling.tail or '') + text
    parent.remove(node)
    return node


def replaceNode(node, *replacements):
    insertBefore(node, *replacements)
    removeNode(node)
    if replacements:
        return replacements[0]


def appendContents(el, container):
    # Accepts either an iterable *or* a container element
    if isElement(container):
        container = childNodes(container, clear=True)
    appendChild(el, *container)
    return el


def replaceContents(el, newElements):
    clearContents(el)
    return appendContents(el, newElements)


def moveContents(toEl, fromEl):
    replaceContents(toEl, fromEl)
    fromEl.text = ''

def wrapContents(parentEl, wrapperEl):
    appendContents(wrapperEl, parentEl)
    appendChild(parentEl, wrapperEl)
    return parentEl


def headingLevelOfElement(el):
    for el in relevantHeadings(el, levels=[2,3,4,5,6]):
        if el.get('data-level') is not None:
            return el.get('data-level')
    return None


def relevantHeadings(startEl, levels=None):
    if levels is None:
        levels = [1,2,3,4,5,6]
    levels = ["h" + str(level) for level in levels]
    currentHeadingLevel = float('inf')
    for el in scopingElements(startEl, *levels):
        tagLevel = int(el.tag[1])
        if tagLevel < currentHeadingLevel:
            yield el
            currentHeadingLevel = tagLevel
        if tagLevel == 2:
            return


def sectionName(el):
    '''
    Return the name of the nearest section to el,
    or None if that section isn't meant to be referenced.
    '''
    h = nextIter(relevantHeadings(el))
    if h is None:
        return "Unnamed section"
    if hasClass(h, "no-ref"):
        return None
    return textContent(h)


def scopingElements(startEl, *tags):
    # Elements that could form a "scope" for the startEl
    # Ancestors, and preceding siblings of ancestors.
    # Maps to the things that can establish a counter scope.
    tagFilter = set(tags)

    for el in startEl.itersiblings(preceding=True, *tags):
        yield el
    for el in startEl.iterancestors():
        if el.tag in tagFilter:
            yield el
        for el in el.itersiblings(preceding=True, *tags):
            yield el


def previousElements(startEl, tag=None, *tags):
    # Elements preceding the startEl in document order.
    # Like .iter(), but in the opposite direction.
    els = []
    for el in startEl.getroottree().getroot().iter(tag=tag, *tags):
        if el == startEl:
            return reversed(els)
        els.append(el)
    return els


def childElements(parentEl, tag="*", *tags, **stuff):
    if len(parentEl) == 0:
        return iter(())
    return parentEl.iterchildren(tag=tag, *tags, **stuff)


def childNodes(parentEl, clear=False, skipOddNodes=True):
    '''
    This function returns all the nodes in a parent element in the DOM sense,
    mixing text nodes (strings) and other nodes together
    (rather than LXML's default stupid handling of text).

    If you set "clear" to True, it'll
    1. remove all of parentEl's children,
       so you can append nodes back to it safely, and
    2. Set parentEl.text and child elements' .tail to null,
       again so you can safely append text to parentEl.
    In other words, the following is a no-op:

    ```
    appendChild(parentEl, *childNodes(parentEl, clear=True))
    ```

    Using clear=True is required if you're going to be modifying the element or its children,
    otherwise you'll get weird results (repeated/misplaced text).
    But if you're just reading nodes,
    it's not necessary.

    skipOddNodes ensures that the return value will only be text and Element nodes;
    if it's false, there might be comments, PIs, etc.
    '''
    if isinstance(parentEl, list):
        return parentEl
    ret = []
    if not emptyText(parentEl.text, wsAllowed=False):
        ret.append(parentEl.text)
        if clear:
            parentEl.text = None
    for c in childElements(parentEl, tag=None):
        if skipOddNodes and isOddNode(c):
            pass
        else:
            ret.append(c)
        if not emptyText(c.tail, wsAllowed=False):
            ret.append(c.tail)
            if clear:
                c.tail = None
    if clear:
        clearContents(parentEl)
    return ret


def nodeIter(el, clear=False, skipOddNodes=True):
    # Iterates thru an element and all its descendants,
    # yielding up each child node it sees in depth-first order.
    # (In other words, same as el.iter(),
    #  but returning nodes+strings rather than the stupid LXML model.)
    # Takes the same kwargs as childNodes
    if isinstance(el, str):
        yield el
        return
    if isinstance(el, etree._ElementTree):
        el = el.getroot()
    text = el.text
    tail = el.tail
    if clear:
        el.text = None
        el.tail = None
    yield el
    if text is not None:
        yield text
    for c in childElements(el, tag=None):
        if skipOddNodes and isOddNode(c):
            continue
        # yield from nodeIter(c, clear=clear, skipOddNodes=skipOddNodes)
        for grandChild in nodeIter(c, clear=clear, skipOddNodes=skipOddNodes):
            yield grandChild
    if tail is not None:
        yield tail


def treeAttr(el, attrName):
    # Find the nearest instance of the given attr in the tree
    # Useful for when you can put an attr on an ancestor and apply it to all contents.
    # Returns attrValue or None if nothing is found.
    import itertools as it
    for target in it.chain([el], el.iterancestors()):
        if target.get(attrName) is not None:
            return target.get(attrName)


def closestAttr(el, *attrs):
    # Like treeAttr, but can provide multiple attr names, and returns the first one found.
    # Useful with combos like highlight/nohighlight
    # If multiple target attrs show up on same element, priority is calling order.
    # Returns a tuple of (attrName, attrValue) or (None, None) if nothing is found.
    import itertools as it
    for target in it.chain([el], el.iterancestors()):
        for attrName in attrs:
            if target.get(attrName) is not None:
                return attrName, target.get(attrName)
    return None, None


def closestAncestor(el, pred):
    # Finds the nearest ancestor matching a predicate
    for target in el.iterancestors():
        if pred(target):
            return target


def filterAncestors(el, pred):
    # Returns all ancestors that match the predicate
    for target in el.iterancestors():
        if pred(target):
            yield target


def hasAncestor(el, pred):
    return closestAncestor(el, pred) is not None


def removeAttr(el, *attrNames):
    # Remove an attribute, silently ignoring if attr doesn't exist.
    for attrName in attrNames:
        if attrName in el.attrib:
            del el.attrib[attrName]
    return el


def hasAttrs(el):
    return bool(el.attrib)


def addClass(el, cls):
    if el.get('class') is None:
        el.set('class', cls)
    elif hasClass(el, cls):
        pass
    else:
        el.set('class', "{0} {1}".format(el.get('class'), cls))


_classMap = {}
def hasClass(el, cls, classMap=_classMap):
    elClass = el.get("class")
    if elClass is None:
        return False
    if cls not in elClass:
        return False
    key = cls, elClass
    if key in classMap:
        return classMap[key]
    ret = re.search(r"(^|\s)"+cls+r"($|\s)", elClass)
    classMap[key] = ret
    return ret


def removeClass(el, cls):
    oldClass = el.get('class')
    if oldClass is None:
        return
    newClass = ' '.join(c for c in oldClass.split() if c != cls)
    if newClass == "":
        del el.attrib['class']
    else:
        el.set('class', newClass)


def isElement(node):
    # LXML HAS THE DUMBEST XML TREE DATA MODEL IN THE WORLD
    return etree.iselement(node) and isinstance(node.tag, str)


def isOddNode(node):
    # Something other than an element node or string.
    if isinstance(node, str):
        return False
    if isElement(node):
        return False
    return True


def isNormative(el, doc):
    # Returns whether the element is "informative" or "normative" with a crude algo.
    # Currently just tests whether the element is in a class=example or class=note block, or not.
    if el in _normativeElCache:
        return _normativeElCache[el]
    informativeClasses = ["note", "example", "non-normative", "informative"] + doc.md.informativeClasses
    for cls in informativeClasses:
        if hasClass(el, cls):
            _normativeElCache[el] = False
            return False
    if hasClass(el, "normative"):
        _normativeElCache[el] = True
        return True
    parent = parentElement(el)
    if not isElement(parent):
        # Went past the root without finding any indicator,
        # so normative by default.
        _normativeElCache[el] = True
        return True
    # Otherwise, walk the tree
    norm = isNormative(parent, doc)
    _normativeElCache[el] = norm
    return norm
_normativeElCache = {}


def isEmpty(el):
    # Returns whether the element is empty - no text or children.
    return (el.text is None or el.text.strip() == "") and len(el) == 0


def hasChildElements(el):
    try:
        next(childElements(el))
        return True
    except StopIteration:
        return False


# If the element has one child element, returns it.
# Otherwise, returns None
def hasOnlyChild(el):
    if not emptyText(el.text):
        # Has significant child text
        return None
    children = childElements(el)
    single = next(children, None)
    if single is None:
        # No children
        return None
    if not emptyText(single.tail):
        # Has significant child text following the child element
        return None
    if next(children, None) is not None:
        # At least two children
        return None
    return single


def fixTypography(text):
    # Replace straight aposes with curly quotes for possessives and contractions.
    text = re.sub(r"([\w])'([\w])", r"\1’\2", text)
    text = re.sub(r"(</[\w]+>)'([\w])", r"\1’\2", text)
    # Fix line-ending em dashes, or --, by moving the previous line up, so no space.
    text = re.sub(r"([^<][^!])(—|--)\r?\n\s*(\S)", r"\1—<wbr>\3", text)
    return text


def fixSurroundingTypography(el):
    # Applies some of the fixTypography changes to the content surrounding an element.
    # Used when a shorthand prevented fixTypography from firing previously.
    if el.tail is not None and el.tail.startswith("'"):
        el.tail = "’" + el.tail[1:]
    return el


def unfixTypography(text):
    # Replace curly quotes with straight quotes, and emdashes with double dashes.
    text = re.sub(r"’", r"'", text)
    # Fix line-ending em dashes, or --, by moving the previous line up, so no space.
    text = re.sub(r"—<wbr>", r"--", text)
    return text


def emptyText(text, wsAllowed=True):
    # Because LXML represents a complete lack of text as None,
    # you can't do something like `el.text.strip() == ""` to test for emptiness.
    # wsAllowed controls whether whitespace-only strings count as empty or not
    if text is None:
        return True
    if not wsAllowed:
        return text == ""
    return text.strip() == ""


def hashContents(el):
    # Hash the contents of an element into an 8-character alphanum string.
    # Generally used for generating probably-unique IDs.
    # Normalize whitespace away to avoid git-related newline normalization issues.
    text = re.sub(r"\s+", " ", textContent(el).strip()).encode("ascii", "xmlcharrefreplace")
    return hashlib.md5(text).hexdigest()[0:8]


def replaceMacros(text, macros):
    # `macros` is a dict of {lowercaseMacroName => replacementText}
    # Macro syntax is [FOO], where FOO is /[A-Z0-9-]+/
    # If written as [FOO?], failure to find a matching macro just replaced it with nothing;
    # otherwise, it throws a fatal error.
    def macroReplacer(match):
        fullText = match.group(0)
        innerText = match.group(2).lower() or ""
        optional = match.group(3) == "?"
        if fullText.startswith("\\"):
            # Escaped
            return fullText[1:]
        if fullText.startswith("[["):
            # Actually a biblio link
            return fullText
        if re.match("[\d-]+$", innerText):
            # No refs are all-digits (this is probably JS code, or a regex/grammar).
            return fullText
        if innerText in macros:
            # For some reason I store all the macros in lowercase,
            # despite requiring them to be spelled with uppercase.
            return str(macros[innerText])
        # Nothing has matched, so start failing the macros.
        if optional:
            return ""
        die("Found unmatched text macro {0}. Correct the macro, or escape it with a leading backslash.", fullText)
        return fullText
    return re.sub(r"(\\|\[)?\[([A-Z0-9-]+)(\??)\]", macroReplacer, text)


def replaceAwkwardCSSShorthands(text):
    # Replace the <<production>> shortcuts, because they won't survive the HTML parser.
    def replaceProduction(match):
        syntaxAttr = escapeAttr(match.group(0))
        escape, text = match.groups()
        if escape:
            return escapeHTML(match.group(0)[1:])
        return "<fake-production-placeholder class=production bs-autolink-syntax='{0}'>{1}</fake-production-placeholder>".format(syntaxAttr, text)
    text = re.sub(r"(\\)?<<([^>\n]+)>>", replaceProduction, text)

    # Replace the ''maybe link'' shortcuts.
    # They'll survive the HTML parser,
    # but the current shorthand-recognizer code won't find them if they contain an element.
    # (The other shortcuts are "atomic" and can't contain elements.)
    def replaceMaybe(match):
        syntaxAttr = escapeAttr(match.group(0))
        escape, text = match.groups()
        if escape:
            return escapeHTML(match.group(0)[1:])
        return "<fake-maybe-placeholder bs-autolink-syntax='{0}'>{1}</fake-maybe-placeholder>".format(syntaxAttr, text)
    text = re.sub(r"(\\)?''([^=\n]+?)''", replaceMaybe, text)
    return text


def fixupIDs(doc, els):
    addOldIDs(els)
    dedupIDs(doc)


def safeID(transOrDoc, id):
    if isinstance(transOrDoc, dict):
        trans = transOrDoc
    else:
        trans = transOrDoc.md.translateIDs
    if id in trans:
        return trans[id]
    return id


def addOldIDs(els):
    for el in els:
        if not el.get("oldids"):
            continue
        oldIDs = [id.strip() for id in el.get("oldids").strip().split(",")]
        for oldID in oldIDs:
            appendChild(el, E.span({"id":oldID}))
        removeAttr(el, "oldids")


def dedupIDs(doc):
    import itertools as iter

    def findId(id):
        return find("#" + id, doc) is not None
    ids = DefaultOrderedDict(list)
    for el in findAll("[id]", doc):
        ids[el.get('id')].append(el)
    for dupeId,els in list(ids.items()):
        if len(els) < 2:
            # Only one instance, so nothing to do.
            continue
        warnAboutDupes = True
        if re.match(r"issue-[0-9a-fA-F]{8}$", dupeId):
            # Don't warn about issues, it's okay if they have the same ID because they're identical text.
            warnAboutDupes = False
        ints = iter.count(1)
        for el in els[1:]:
            # If I registered an alternate ID, try to use that.
            if el.get('data-alternate-id'):
                altId = el.get('data-alternate-id')
                if altId not in ids:
                    el.set("id", safeID(doc, el.get("data-alternate-id")))
                    ids[altId].append(el)
                    continue
            if el.get("data-silently-dedup") is not None:
                warnAboutDupes = False
            if dupeId.startswith("ref-for-"):
                warnAboutDupes = False
            # Try to de-dup the id by appending an integer after it.
            if warnAboutDupes:
                warn("Multiple elements have the same ID '{0}'.\nDeduping, but this ID may not be stable across revisions.", dupeId, el=el)
            for x in ints:
                altId = "{0}{1}".format(dupeId, circledDigits(x))
                if altId not in ids:
                    el.set("id", safeID(doc, altId))
                    ids[altId].append(el)
                    break


def approximateLineNumber(el, setIntermediate=True):
    if el.get('line-number'):
        return el.get('line-number')
    parent = parentElement(el)
    if not isElement(parent):
        if el.tag == "html":
            return None
        return None
    approx = approximateLineNumber(parent, setIntermediate=setIntermediate)
    if approx is None:
        return None
    if approx[0].isdigit():
        approx = "~" + approx
    if setIntermediate:
        el.set('line-number', approx)
    return approx


def circledDigits(num):
    '''
    Converts a base-10 number into a string using unicode circled digits.
    That is, 123 becomes "①②③"
    '''
    num = int(num)
    assert(num >= 0)
    digits = ["⓪","①","②","③","④","⑤","⑥","⑦","⑧","⑨"]
    result = "".join(digits[int(d)] for d in str(num))
    return result


def nextIter(it, default=None):
    '''
    Returns the next element of the iterator,
    returning the default value if it's empty,
    rather than throwing an error.
    '''
    try:
        return next(iter(it))
    except StopIteration:
        return default


def createElement(tag, attrs={}, *children):
    el = etree.Element(tag, {n:v for n,v in attrs.items() if v is not None})
    for child in children:
        appendChild(el, child)
    return el


class ElementCreationHelper:
    def __getattr__(self, name):
        def _creater(*children):
            if children and not (isinstance(children[0], str) or isElement(children[0])):
                attrs = children[0]
                children = children[1:]
            else:
                attrs = {}
            return createElement(name, attrs, *children)
        return _creater
E = ElementCreationHelper()



from . import railroaddiagrams as rr
from .messages import *


def parse(string):
    '''
    Parses a DSL for railroad diagrams,
    based on significant whitespace.
    Each command must be on its own line,
    and is written like "Sequence:\n".
    Children are indented on following lines.
    Some commands have non-child arguments;
    for block commands they are put on the same line, after the :,
    for text commands they are put on the same line *before* the :,
    like:
        Choice: 0
            Terminal: foo
            Terminal raw: bar
    '''
    import re

    lines = string.splitlines()

    # Strip off any common initial whitespace from lines.
    initialIndent = re.match(r"(\s*)", lines[0]).group(1)
    for i, line in enumerate(lines):
        if line.startswith(initialIndent):
            lines[i] = line[len(initialIndent):]
        else:
            die("Inconsistent indentation: line {0} is indented less than the first line.", i)
            return rr.Diagram()

    # Determine subsequent indentation
    for line in lines:
        match = re.match(r"(\s+)", line)
        if match:
            indentText = match.group(1)
            break
    else:
        indentText = "\t"

    # Turn lines into tree
    lastIndent = 0
    tree = {"command":"Diagram", "prelude": "", "children": []}
    activeCommands = {"0": tree}
    blockNames = "And|Seq|Sequence|Stack|Or|Choice|Opt|Optional|Plus|OneOrMore|Star|ZeroOrMore"
    textNames = "T|Terminal|N|NonTerminal|C|Comment|S|Skip"
    for i, line in enumerate(lines, 1):
        indent = 0
        while line.startswith(indentText):
            indent += 1
            line = line[len(indentText):]
        if indent > lastIndent + 1:
            die("Line {0} jumps more than 1 indent level from the previous line:\n{1}", i, line.strip())
            return rr.Diagram()
        lastIndent = indent
        if re.match(r"\s*({0})\W".format(blockNames), line):
            match = re.match(r"\s*(\w+)\s*:\s*(.*)", line)
            if not match:
                die("Line {0} doesn't match the grammar 'Command: optional-prelude'. Got:\n{1}", i, line.strip())
                return rr.Diagram()
            command = match.group(1)
            prelude = match.group(2).strip()
            node = {"command": command, "prelude": prelude, "children": [], "line": i}
        elif re.match(r"\s*({0})\W".format(textNames), line):
            match = re.match(r"\s*(\w+)(\s[\w\s]+)?:\s*(.*)", line)
            if not match:
                die("Line {0} doesn't match the grammar 'Command [optional prelude]: text'. Got:\n{1},", i, line.strip())
                return rr.Diagram()
            command = match.group(1)
            if match.group(2):
                prelude = match.group(2).strip()
            else:
                prelude = None
            text = match.group(3).strip()
            node = {"command": command, "prelude": prelude, "text":text, "children": [], "line": i}
        else:
            die("Line {0} doesn't contain a valid railroad-diagram command. Got:\n{1}", i, line.strip())
            return

        activeCommands[str(indent)]['children'].append(node)
        activeCommands[str(indent + 1)] = node

    return _createDiagram(**tree)


def _createDiagram(command, prelude, children, text=None, line=-1):
    '''
    From a tree of commands,
    create an actual Diagram class.
    Each command must be {command, prelude, children}
    '''
    if command == "Diagram":
        children = [_f for _f in [_createDiagram(**child) for child in children] if _f]
        return rr.Diagram(*children)
    elif command in ("T", "Terminal"):
        if children:
            return die("Line {0} - Terminal commands cannot have children.", line)
        return rr.Terminal(text, prelude)
    elif command in ("N", "NonTerminal"):
        if children:
            return die("Line {0} - NonTerminal commands cannot have children.", line)
        return rr.NonTerminal(text, prelude)
    elif command in ("C", "Comment"):
        if children:
            return die("Line {0} - Comment commands cannot have children.", line)
        return rr.Comment(text, prelude)
    elif command in ("S", "Skip"):
        if children:
            return die("Line {0} - Skip commands cannot have children.", line)
        if text:
            return die("Line {0} - Skip commands cannot have text.", line)
        return rr.Skip()
    elif command in ("And", "Seq", "Sequence"):
        if prelude:
            return die("Line {0} - Sequence commands cannot have preludes.", line)
        if not children:
            return die("Line {0} - Sequence commands need at least one child.", line)
        children = [_f for _f in [_createDiagram(**child) for child in children] if _f]
        return rr.Sequence(*children)
    elif command in ("Stack", ):
        if prelude:
            return die("Line {0} - Stack commands cannot have preludes.", line)
        if not children:
            return die("Line {0} - Stack commands need at least one child.", line)
        children = [_f for _f in [_createDiagram(**child) for child in children] if _f]
        return rr.Stack(*children)
    elif command in ("Or", "Choice"):
        if prelude == "":
            default = 0
        else:
            try:
                default = int(prelude)
            except:
                die("Line {0} - Choice preludes must be an integer. Got:\n{1}", line, prelude)
                default = 0
        if not children:
            return die("Line {0} - Choice commands need at least one child.", line)
        children = [_f for _f in [_createDiagram(**child) for child in children] if _f]
        return rr.Choice(default, *children)
    elif command in ("Opt", "Optional"):
        if prelude not in ("", "skip"):
            return die("Line {0} - Optional preludes must be nothing or 'skip'. Got:\n{1}", line, prelude)
        if len(children) != 1:
            return die("Line {0} - Optional commands need exactly one child.", line)
        children = [_f for _f in [_createDiagram(**child) for child in children] if _f]
        return rr.Optional(*children, skip=(prelude == "skip"))
    elif command in ("Plus", "OneOrMore"):
        if prelude:
            return die("Line {0} - OneOrMore commands cannot have preludes.", line)
        if 0 == len(children) > 2:
            return die("Line {0} - OneOrMore commands must have one or two children.", line)
        children = [_f for _f in [_createDiagram(**child) for child in children] if _f]
        return rr.OneOrMore(*children)
    elif command in ("Star", "ZeroOrMore"):
        if prelude:
            return die("Line {0} - ZeroOrMore commands cannot have preludes.", line)
        if 0 == len(children) > 2:
            return die("Line {0} - ZeroOrMore commands must have one or two children.", line)
        children = [_f for _f in [_createDiagram(**child) for child in children] if _f]
        return rr.ZeroOrMore(*children)
    else:
        return die("Line {0} - Unknown command '{1}'.", line, command)



# -*- coding: utf-8 -*-


from ..messages import printColor
from ..html import outerHTML, isElement

def printjson(x, indent=2, level=0):
    if isinstance(indent, int):
        # Convert into a number of spaces.
        indent = " " * indent
    x = getjson(x)
    if isinstance(x, dict):
        ret = printjsonobject(x, indent, level)
    elif isinstance(x, list):
        if len(x) > 0 and isinstance(getjson(x[0]), dict):
            ret = printjsonobjectarray(x, indent, level)
        else:
            ret = printjsonsimplearray(x, indent, level)
    else:
        ret = printjsonprimitive(x)
    if level == 0 and ret.startswith("\n"):
        ret = ret[1:]
    return ret
    #return json.dumps(obj, indent=2, default=lambda x:x.__json__())


def getjson(x):
    try:
        return x.__json__()
    except AttributeError:
        return x


def printjsonobject(x, indent, level):
    x = getjson(x)
    ret = ""
    maxKeyLength = 0
    for k in x.keys():
        maxKeyLength = max(maxKeyLength, len(k))
    for k,v in x.items():
        ret += "\n" + (indent * level) + printColor((k + ": ").ljust(maxKeyLength + 2), "cyan") + printjson(v, indent, level + 1)
    return ret


def printjsonobjectarray(x, indent, level):
    # Prints an array of objects
    x = getjson(x)
    ret = ""
    for i,v in enumerate(x):
        if i != 0:
            ret += "\n" + (indent * level) + printColor("=" * 10, "blue")
        ret += printjsonobject(v, indent, level)
    return ret


def printjsonsimplearray(x, indent, level):
    x = getjson(x)
    ret = printColor("[", "blue")
    for i,v in enumerate(x):
        if i != 0:
            ret += printColor(", ", "blue")
        ret += printjsonprimitive(v)
    ret += printColor("]", "blue")
    return ret


def printjsonprimitive(x):
    x = getjson(x)
    if isinstance(x, int):
        return str(x)
    if isinstance(x, str):
        return x
    if isinstance(x, bool):
        return str(x)
    if x is None:
        return "null"
    if isElement(x):
        return repr(x) + ":" + outerHTML(x)
    raise Exception(x)



<<<<<<< HEAD
import re
from ..html import E, outerHTML
from . import steps
from ..messages import die

class PropdescShorthand:
    def __init__(self):
        self.stage = "start"
        self.escapedText = None
        self.linkText = []
        self.bsAutolink = ""

    def respond(self, match, dom=None):
        if self.stage == "start":
            return self.respondStart(match)
        elif self.stage == "link text":
            return self.respondLinkText(match, dom)
        elif self.stage == "end":
            return self.respondEnd()

    def respondStart(self, match):
        self.bsAutolink = match.group(0)
        escape, self.linkFor, self.lt, self.linkType, hasLinkText = match.groups()
        if escape:
            self.escapedText = match.group(0)[1:]

        if match.groups(0) == "'-":
            # Not a valid property actually.
            return steps.Failure()

        if self.linkFor == "":
            self.linkFor = "/"

        if self.linkType is None:
            self.linkType = "propdesc"

        if hasLinkText:
            self.stage = "link text"
            return steps.NextBody(endRe)
        else:
            self.stage = "end"
            return steps.NextLiteral(endRe)

    def respondLinkText(self, match, dom):
        self.linkText = dom
        self.bsAutolink += outerHTML(dom)
        return self.respondEnd()

    def respondEnd(self):
        if self.escapedText:
            return steps.Success(skips=["'"], nodes=[self.escapedText[1:], *self.linkText, "'"])

        self.bsAutolink += "'"

        if self.linkType not in ["property", "descriptor", "propdesc"]:
            die("Shorthand {0} gives type as '{1}', but only 'property' and 'descriptor' are allowed.", self.bsAutolink, linkType)
            return steps.Success(E.span(self.bsAutolink))

        if self.linkText is None:
            self.linkText = lt

        attrs = {
            "data-link-type":self.linkType,
            "class":"property",
            "for": self.linkFor,
            "lt": self.lt,
            "bs-autolink-syntax":self.bsAutolink
            }
        return steps.Success(E.a(attrs, linkText))


PropdescShorthand.startRe = re.compile(r"""
                        (\\)?
                        '
                        (?:([^\s'|]*)/)?
                        ([\w*-]+)
                        (?:!!([\w-]+))?
                        (\|)?
                        """, re.X)

endRe = re.compile("'")



# -*- coding: utf-8 -*-


from .DefaultOrderedDict import DefaultOrderedDict
from .html import *

def addDfnPanels(doc, dfns):
    # Constructs "dfn panels" which show all the local references to a term
    atLeastOnePanel = False
    # Gather all the <a href>s together
    allRefs = DefaultOrderedDict(list)
    for a in findAll("a", doc):
        href = a.get("href")
        if href is None:
            continue
        if not href.startswith("#"):
            continue
        allRefs[href[1:]].append(a)
    for dfn in dfns:
        id = dfn.get("id")
        if not id:
            # Something went wrong, bail.
            continue
        refs = DefaultOrderedDict(list)
        for link in allRefs[id]:
            section = sectionName(link)
            if section is not None:
                refs[section].append(link)
        if not refs:
            # Just insert a self-link instead
            appendChild(dfn,
                        E.a({"href": "#" + escapeUrlFrag(id), "class":"self-link"}))
            continue
        addClass(dfn, "dfn-paneled")
        atLeastOnePanel = True
        panel = E.aside({"class": "dfn-panel", "data-for": id},
                        E.b(E.a({"href":"#" + escapeUrlFrag(id)}, "#" + id)),
                        E.b("Referenced in:"))
        ul = appendChild(panel, E.ul())
        for text,els in refs.items():
            li = appendChild(ul, E.li())
            for i,el in enumerate(els):
                refID = el.get("id")
                if refID is None:
                    refID = "ref-for-{0}".format(id)
                    el.set("id", safeID(doc, refID))
                if i == 0:
                    appendChild(li,
                                E.a({"href": "#" + escapeUrlFrag(refID), "data-silently-dedup": ""}, text))
                else:
                    appendChild(li,
                                " ",
                                E.a({"href": "#" + escapeUrlFrag(refID), "data-silently-dedup": ""}, "(" + str(i + 1) + ")"))
        appendChild(doc.body, panel)
    if atLeastOnePanel:
        doc.extraScripts['script-dfn-panel'] = dfnPanelScript
        doc.extraStyles['style-dfn-panel'] = dfnPanelStyle

def addExternalDfnPanel(termEl, ref, elsFromHref, doc):
    # Constructs "dfn panels" which show all the local references to an external term
    # Gather all the <a href>s together
    refs = DefaultOrderedDict(list)
    for el in elsFromHref[ref.url]:
        section = sectionName(el) or "Unnumbered Section"
        refs[section].append(el)
    if len(refs):
        addClass(termEl, "dfn-paneled")
        _,_,refID = ref.url.partition("#")
        termID = "term-for-{0}".format(refID)
        termEl.set("id", termID)
        termEl.set("data-silently-dedup", "")
        panel = E.aside({"class": "dfn-panel", "data-for": termID},
                        E.a({"href":ref.url}, ref.url),
                        E.b("Referenced in:"))
        ul = appendChild(panel, E.ul())
        for text,els in refs.items():
            li = appendChild(ul, E.li())
            for i,el in enumerate(els):
                linkID = el.get("id")
                if linkID is None:
                    linkID = "termref-for-".format(refID)
                    el.set("id", safeID(doc, linkID))
                    el.set("data-silently-dedup", "")
                if i == 0:
                    appendChild(li,
                                E.a({"href": "#" + escapeUrlFrag(linkID), "data-silently-dedup": ""}, text))
                else:
                    appendChild(li,
                                " ",
                                E.a({"href": "#" + escapeUrlFrag(linkID), "data-silently-dedup": ""}, "(" + str(i + 1) + ")"))
        appendChild(doc.body, panel)
        doc.extraScripts['script-dfn-panel'] = dfnPanelScript
        doc.extraStyles['style-dfn-panel'] = dfnPanelStyle


dfnPanelScript = '''
document.body.addEventListener("click", function(e) {
    var queryAll = function(sel) { return [].slice.call(document.querySelectorAll(sel)); }
    // Find the dfn element or panel, if any, that was clicked on.
    var el = e.target;
    var target;
    var hitALink = false;
    while(el.parentElement) {
        if(el.tagName == "A") {
            // Clicking on a link in a <dfn> shouldn't summon the panel
            hitALink = true;
        }
        if(el.classList.contains("dfn-paneled")) {
            target = "dfn";
            break;
        }
        if(el.classList.contains("dfn-panel")) {
            target = "dfn-panel";
            break;
        }
        el = el.parentElement;
    }
    if(target != "dfn-panel") {
        // Turn off any currently "on" or "activated" panels.
        queryAll(".dfn-panel.on, .dfn-panel.activated").forEach(function(el){
            el.classList.remove("on");
            el.classList.remove("activated");
        });
    }
    if(target == "dfn" && !hitALink) {
        // open the panel
        var dfnPanel = document.querySelector(".dfn-panel[data-for='" + el.id + "']");
        if(dfnPanel) {
            dfnPanel.classList.add("on");
            var rect = el.getBoundingClientRect();
            dfnPanel.style.left = window.scrollX + rect.right + 5 + "px";
            dfnPanel.style.top = window.scrollY + rect.top + "px";
            var panelRect = dfnPanel.getBoundingClientRect();
            var panelWidth = panelRect.right - panelRect.left;
            if(panelRect.right > document.body.scrollWidth && (rect.left - (panelWidth + 5)) > 0) {
                // Reposition, because the panel is overflowing
                dfnPanel.style.left = window.scrollX + rect.left - (panelWidth + 5) + "px";
            }
        } else {
            console.log("Couldn't find .dfn-panel[data-for='" + el.id + "']");
        }
    } else if(target == "dfn-panel") {
        // Switch it to "activated" state, which pins it.
        el.classList.add("activated");
        el.style.left = null;
        el.style.top = null;
    }

});
'''

dfnPanelStyle = '''
.dfn-panel {
    position: absolute;
    z-index: 35;
    height: auto;
    width: -webkit-fit-content;
    width: fit-content;
    max-width: 300px;
    max-height: 500px;
    overflow: auto;
    padding: 0.5em 0.75em;
    font: small Helvetica Neue, sans-serif, Droid Sans Fallback;
    background: #DDDDDD;
    color: black;
    border: outset 0.2em;
}
.dfn-panel:not(.on) { display: none; }
.dfn-panel * { margin: 0; padding: 0; text-indent: 0; }
.dfn-panel > b { display: block; }
.dfn-panel a { color: black; }
.dfn-panel a:not(:hover) { text-decoration: none !important; border-bottom: none !important; }
.dfn-panel > b + b { margin-top: 0.25em; }
.dfn-panel ul { padding: 0; }
.dfn-panel li { list-style: inside; }
.dfn-panel.activated {
    display: inline-block;
    position: fixed;
    left: .5em;
    bottom: 2em;
    margin: 0 auto;
    max-width: calc(100vw - 1.5em - .4em - .5em);
    max-height: 30vh;
}

.dfn-paneled { cursor: pointer; }
'''



# -*- coding: utf-8 -*-


from .main import englishFromList
from ..messages import *

shortToLongStatus = {
    "DREAM": "A Collection of Interesting Ideas",
    "LS": "Living Standard",
    "LS-COMMIT": "Commit Snapshot",
    "LS-BRANCH": "Branch Snapshot",
    "LS-PR": "PR Preview",
    "LD": "Living Document",
    "DRAFT-FINDING": "Draft Finding",
    "FINDING": "Finding",
    "whatwg/RD": "Review Draft",
    "w3c/ED": "Editor's Draft",
    "w3c/WD": "W3C Working Draft",
    "w3c/FPWD": "W3C First Public Working Draft",
    "w3c/LCWD": "W3C Last Call Working Draft",
    "w3c/CR": "W3C Candidate Recommendation",
    "w3c/PR": "W3C Proposed Recommendation",
    "w3c/REC": "W3C Recommendation",
    "w3c/PER": "W3C Proposed Edited Recommendation",
    "w3c/WG-NOTE": "W3C Working Group Note",
    "w3c/IG-NOTE": "W3C Interest Group Note",
    "w3c/NOTE": "W3C Note",
    "w3c/MO": "W3C Member-only Draft",
    "w3c/UD": "Unofficial Proposal Draft",
    "w3c/CG-DRAFT": "Draft Community Group Report",
    "w3c/CG-FINAL": "Final Community Group Report",
    "tc39/STAGE0": "Stage 0: Strawman",
    "tc39/STAGE1": "Stage 1: Proposal",
    "tc39/STAGE2": "Stage 2: Draft",
    "tc39/STAGE3": "Stage 3: Candidate",
    "tc39/STAGE4": "Stage 4: Finished",
    "iso/I": "Issue",
    "iso/DR": "Defect Report",
    "iso/D": "Draft Proposal",
    "iso/P": "Published Proposal",
    "iso/MEET": "Meeting Announcements",
    "iso/RESP": "Records of Response",
    "iso/MIN": "Minutes",
    "iso/ER": "Editor's Report",
    "iso/SD": "Standing Document",
    "iso/PWI": "Preliminary Work Item",
    "iso/NP": "New Proposal",
    "iso/NWIP": "New Work Item Proposal",
    "iso/WD": "Working Draft",
    "iso/CD": "Committee Draft",
    "iso/FCD": "Final Committee Draft",
    "iso/DIS": "Draft International Standard",
    "iso/FDIS": "Final Draft International Standard",
    "iso/PRF": "Proof of a new International Standard",
    "iso/IS": "International Standard",
    "iso/TR": "Technical Report",
    "iso/DTR": "Draft Technical Report",
    "iso/TS": "Technical Specification",
    "iso/DTS": "Draft Technical Specification",
    "iso/PAS": "Publicly Available Specification",
    "iso/TTA": "Technology Trends Assessment",
    "iso/IWA": "International Workshop Agreement",
    "iso/COR": "Technical Corrigendum",
    "iso/GUIDE": "Guidance to Technical Committees",
    "iso/NP-AMD": "New Proposal Amendment",
    "iso/AWI-AMD": "Approved new Work Item Amendment",
    "iso/WD-AMD": "Working Draft Amendment",
    "iso/CD-AMD": "Committee Draft Amendment",
    "iso/PD-AMD": "Proposed Draft Amendment",
    "iso/FPD-AMD": "Final Proposed Draft Amendment",
    "iso/D-AMD": "Draft Amendment",
    "iso/FD-AMD": "Final Draft Amendment",
    "iso/PRF-AMD": "Proof Amendment",
    "iso/AMD": "Amendment",
    "fido/ED": "Editor's Draft",
    "fido/WD": "Working Draft",
    "fido/RD": "Review Draft",
    "fido/ID": "Implementation Draft",
    "fido/PS": "Proposed Standard",
    "fido/FD": "Final Document",
    "khronos/ED": "Editor's Draft"
}
snapshotStatuses = ["w3c/WD", "w3c/FPWD", "w3c/LCWD", "w3c/CR", "w3c/PR", "w3c/REC", "w3c/PER", "w3c/WG-NOTE", "w3c/IG-NOTE", "w3c/NOTE", "w3c/MO"]
datedStatuses = ["w3c/WD", "w3c/FPWD", "w3c/LCWD", "w3c/CR", "w3c/PR", "w3c/REC", "w3c/PER", "w3c/WG-NOTE", "w3c/IG-NOTE", "w3c/NOTE", "w3c/MO", "whatwg/RD"]
unlevelledStatuses = ["LS", "LD", "DREAM", "w3c/UD", "LS-COMMIT", "LS-BRANCH", "LS-PR", "FINDING", "DRAFT-FINDING", "whatwg/RD"]
deadlineStatuses = ["w3c/LCWD", "w3c/PR"]
noEDStatuses = ["LS", "LS-COMMIT", "LS-BRANCH", "LS-PR", "LD", "FINDING", "DRAFT-FINDING", "DREAM", "iso/NP", "whatwg/RD"]

# W3C statuses are restricted in various confusing ways.

# These statuses are usable by any group operating under the W3C Process
# Document. (So, not by Community and Business Groups.)
w3cProcessDocumentStatuses = frozenset([
    "w3c/ED",
    "w3c/NOTE",
    "w3c/UD"
])

# Interest Groups are limited to these statuses
w3cIGStatuses = frozenset([
    "w3c/IG-NOTE"
]).union(w3cProcessDocumentStatuses)
# Working Groups are limited to these statuses
w3cWGStatuses = frozenset([
    "w3c/WD",
    "w3c/FPWD",
    "w3c/LCWD",
    "w3c/CR",
    "w3c/PR",
    "w3c/REC",
    "w3c/PER",
    "w3c/WG-NOTE"
]).union(w3cProcessDocumentStatuses)
# The TAG is limited to these statuses
w3cTAGStatuses = frozenset([
    "DRAFT-FINDING",
    "FINDING",
    "w3c/WG-NOTE" # despite the TAG not being a WG. I know, it's weird.
]).union(w3cProcessDocumentStatuses)
# Community and Business Groups are limited to these statuses
w3cCommunityStatuses = frozenset([
    "w3c/CG-DRAFT",
    "w3c/CG-FINAL"
])

megaGroups = {
    "w3c": frozenset(["act-framework", "audiowg", "browser-testing-tools", "csswg", "dap", "fxtf", "fxtf-csswg", "geolocation", "houdini", "html", "i18n", "immersivewebcg", "immersivewebwg", "mediacapture", "mediawg", "ping", "privacycg", "processcg", "ricg", "sacg", "secondscreenwg", "serviceworkers", "svg", "tag", "texttracks", "uievents", "wasm", "web-bluetooth-cg", "webapps", "webappsec", "webauthn", "webml", "web-payments", "webperf", "webplatform", "webrtc", "webspecs", "webvr", "wicg"]),
    "whatwg": frozenset(["whatwg"]),
    "tc39": frozenset(["tc39"]),
    "iso": frozenset(["wg14", "wg21"]),
    "fido": frozenset(["fido"]),
    "priv-sec": frozenset(["audiowg", "csswg", "dap", "fxtf", "fxtf-csswg", "geolocation", "houdini", "html", "mediacapture", "mediawg", "ricg", "svg", "texttracks", "uievents", "web-bluetooth-cg", "webappsec", "webplatform", "webspecs", "whatwg"]),
    "khronos": frozenset(["webgl"])
}
# Community and business groups within the W3C:
w3cCgs = frozenset(["immersivewebcg", "privacycg", "processcg", "ricg", "sacg", "web-bluetooth-cg", "wicg"])
assert w3cCgs.issubset(megaGroups["w3c"])
# Interest Groups within the W3C:
w3cIgs = frozenset(["ping"])
assert w3cIgs.issubset(megaGroups["w3c"])

def canonicalizeStatus(rawStatus, group):
    if rawStatus is None:
        return None

    def validateW3Cstatus(group, status, rawStatus):
        if status == "DREAM":
            warn("You used Status: DREAM for a W3C document."
                 + " Consider UD instead.")
            return

        if "w3c/"+status in shortToLongStatus:
            status = "w3c/"+status

        def formatStatusSet(statuses):
            return ", ".join(sorted(set([status.split("/")[-1] for status in statuses])))

        msg = "You used Status: {0}, but {1} limited to these statuses: {2}."

        if group in w3cIgs and status not in w3cIGStatuses:
            warn(msg, rawStatus, "W3C Interest Groups are",
                 formatStatusSet(w3cIGStatuses))

        if group == "tag" and status not in w3cTAGStatuses:
            warn(msg, rawStatus, "the TAG is",
                 formatStatusSet(w3cTAGStatuses))

        if group in w3cCgs and status not in w3cCommunityStatuses:
            warn(msg, rawStatus, "W3C Community and Business Groups are",
                 formatStatusSet(w3cCommunityStatuses))

    def megaGroupsForStatus(status):
        # Returns a list of megagroups that recognize the given status
        megaGroups = []
        for key in shortToLongStatus.keys():
            mg,_,s = key.partition("/")
            if s == status:
                megaGroups.append(mg)
        return megaGroups

    # Canonicalize the rawStatus that was passed in, into a known form.
    # Might be foo/BAR, or just BAR.
    megaGroup,_,status = rawStatus.partition("/")
    if status == "":
        status = megaGroup
        megaGroup = ""
    megaGroup = megaGroup.lower()
    status = status.upper()
    if megaGroup:
        canonStatus = megaGroup + "/" + status
    else:
        canonStatus = status

    if group is not None:
        group = group.lower()

    if group in megaGroups["w3c"]:
        validateW3Cstatus(group, canonStatus, rawStatus)

    # Using a directly-recognized status is A-OK.
    # (Either one of the unrestricted statuses,
    # or one of the restricted statuses with the correct standards-org prefix.)
    if canonStatus in shortToLongStatus:
        return canonStatus

    possibleMgs = megaGroupsForStatus(status)

    # If they specified a standards-org prefix and it wasn't found,
    # that's an error.
    if megaGroup:
        # Was the error because the megagroup doesn't exist?
        if possibleMgs:
            if megaGroup not in megaGroups:
                msg = "Status metadata specified an unrecognized '{0}' organization.".format(megaGroup)
            else:
                msg = "Status '{0}' can't be used with the org '{1}'.".format(status, megaGroup)
            if "" in possibleMgs:
                if len(possibleMgs) == 1:
                    msg += " That status must be used without an org at all, like `Status: {0}`".format(status)
                else:
                    msg += " That status can only be used with the org{0} {1}, or without an org at all.".format(
                        "s" if len(possibleMgs)>1 else "",
                        englishFromList("'{0}'".format(x) for x in possibleMgs if x != ""))
            else:
                if len(possibleMgs) == 1:
                    msg += " That status can only be used with the org '{0}', like `Status: {0}/{1}`".format(possibleMgs[0], status)
                else:
                    msg += " That status can only be used with the orgs {0}.".format(englishFromList("'{0}'".format(x) for x in possibleMgs))

        else:
            if megaGroup not in megaGroups:
                msg = "Unknown Status metadata '{0}'. Check the docs for valid Status values.".format(canonStatus)
            else:
                msg = "Status '{0}' can't be used with the org '{1}'. Check the docs for valid Status values.".format(status, megaGroup)
        die("{0}", msg)
        return canonStatus

    # Otherwise, they provided a bare status.
    # See if their group is compatible with any of the prefixed statuses matching the bare status.
    assert "" not in possibleMgs # if it was here, the literal "in" test would have caught this bare status
    for mg in possibleMgs:
        if group in megaGroups[mg]:
            canonStatus = mg + "/" + status

            if mg == "w3c":
                validateW3Cstatus(group, canonStatus, rawStatus)

            return canonStatus

    # Group isn't in any compatible org, so suggest prefixing.
    if possibleMgs:
        msg = "You used Status: {0}, but that's limited to the {1} org{2}".format(
            rawStatus,
            englishFromList("'{0}'".format(mg) for mg in possibleMgs),
            "s" if len(possibleMgs)>1 else "")
        if group:
            msg += ", and your group '{0}' isn't recognized as being in {1}.".format(group, "any of those orgs" if len(possibleMgs)>1 else "that org")
            msg += " If this is wrong, please file a Bikeshed issue to categorize your group properly, and/or try:\n"
            msg += "\n".join("Status: {0}/{1}".format(mg, status) for mg in possibleMgs)
        else:
            msg += ", and you don't have a Group metadata. Please declare your Group, or check the docs for statuses that can be used by anyone."
    else:
        msg = "Unknown Status metadata '{0}'. Check the docs for valid Status values.".format(canonStatus)
    die("{0}", msg)
    return canonStatus



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()
with open("semver.txt", "r") as fh:
    semver = fh.read().strip()
with open("requirements.txt", "r") as fh:
    install_requires = [x.strip() for x in fh.read().strip().split("\n")]

setup(
    name='bikeshed',
    version=semver,
    author='Tab Atkins-Bittner',
    description="A document-authoring tool mainly intended for web specifications.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/tabatkins/bikeshed/",
    packages=find_packages(),
    include_package_data=True,
    install_requires=install_requires,
    python_requires=">=3.7",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "License :: Public Domain",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.7",
        "Topic :: Software Development :: Documentation",
    ],
    entry_points={'console_scripts': ['bikeshed = bikeshed:main']},
)



# -*- coding: utf-8 -*-

import re
from ..html import *
from ..messages import *

def missingExposed(doc):
    '''
    Checks that every IDL interface or namespace has an [Exposed] attribute.
    Specifically:
    * Any namespace
    * Any non-callback interface without [NoInterfaceObject]
    * Any callback interface that has constants in it.

    Partials never need to declare [Exposed], I think?
    '''
    if not doc.md.complainAbout['missing-exposed']:
        return

    for construct in doc.widl.constructs:
        if construct.extended_attributes is None:
            extendedAttrs = []
        else:
            extendedAttrs = construct.extended_attributes
        if construct.idl_type == "namespace":
            good = False
            for attr in extendedAttrs:
                if attr.name == "Exposed":
                    good = True
                    break
            if not good:
                warn("The '{0}' namespace is missing an [Exposed] extended attribute. Does it need [Exposed=Window], or something more?", construct.name)
        elif construct.idl_type == "interface":
            good = False
            for attr in extendedAttrs:
                if attr.name == "Exposed":
                    good = True
                    break
                if attr.name == "NoInterfaceObject":
                    good = True
                    break
            if not good:
                warn("The '{0}' interface is missing an [Exposed] extended attribute. Does it need [Exposed=Window], or something more?", construct.name)
        elif construct.idl_type == "callback":
            if not hasattr(construct, "interface"):
                # Just a callback function, it's fine
                continue
            for members in construct.interface.members:
                pass



import re
from ..html import E
from . import steps

class VarShorthand:
    def respond(self, match, dom=None):
	escape, text = match.groups()
	    if escape:
	        return steps.Success(skips=["|"], nodes=[match.group(0)[1:]])
    return steps.Success(E.var({"bs-autolink-syntax":match.group(0)}, text)

VarShorthand.startRe = re.compile(r"""
                    (\\)?
                    \|
                    (\w(?:[\w\s-]*\w)?)
                    \|""", re.X)


# -*- coding: utf-8 -*-

import io
import os
import requests

from ..messages import *

def update(path, dryRun=False):
    try:
        say("Downloading web-platform-tests data...")
        response = requests.get("https://wpt.fyi/api/manifest")
        sha = response.headers["x-wpt-sha"]
        jsonData = response.json()
    except Exception as e:
        die("Couldn't download web-platform-tests data.\n{0}", e)
        return

    if "version" not in jsonData:
        die("Can't figure out the WPT data version. Please report this to the maintainer!")
        return

    if jsonData["version"] != 8:
        die("Bikeshed currently only knows how to handle WPT v8 manifest data, but got v{0}. Please report this to the maintainer!", jsonData["version"])
        return

    paths = []
    for testType, typePaths in jsonData["items"].items():
        if testType not in ("manual", "reftest", "testharness", "wdspec"):
            # Not tests
            continue
        collectPaths(paths, typePaths, testType + " ")

    if not dryRun:
        try:
            with io.open(os.path.join(path, "wpt-tests.txt"), 'w', encoding="utf-8") as f:
                f.write("sha: {0}\n".format(sha))
                for path in sorted(paths):
                    f.write(path+"\n")
        except Exception as e:
            die("Couldn't save web-platform-tests data to disk.\n{0}", e)
            return
    say("Success!")


def collectPaths(pathListSoFar, pathTrie, pathPrefix):
    for k,v in pathTrie.items():
        if isinstance(v, dict):
            collectPaths(pathListSoFar, v, "{0}{1}/".format(pathPrefix, k))
        else:
            pathListSoFar.append(pathPrefix+k)
    return pathListSoFar


import re
from ..html import E, outerHTML
from . import steps

class DfnShorthand:
    def __init__(self):
        self.stage = "start"
        self.escapedText = None
        self.linkText = []
        self.bsAutolink = ""

    def respond(self, match, dom=None):
        if self.stage == "start":
            return self.respondStart(match)
        elif self.stage == "link text":
            return self.respondLinkText(match, dom)
        elif self.stage == "end":
            return self.respondEnd()

    def respondStart(self, match):
        self.bsAutolink = match.group(0)
        escape, self.linkFor, self.lt, hasLinkText = match.groups()
        if escape:
            self.escapedText = match.group(0)[1:]

        if self.linkFor == "":
            self.linkFor = "/"

        if hasLinkText:
            self.stage = "link text"
            return steps.NextBody(endRe)
        else:
            self.stage = "end"
            return steps.NextLiteral(endRe)

    def respondLinkText(self, match, dom):
        self.linkText = dom
        self.bsAutolink += outerHTML(dom)
        return self.respondEnd()

    def respondEnd(self):
        if self.escapedText:
            return steps.Success(skips=["["], nodes=[self.escapedText[1:], *self.linkText, "=]"])

        self.bsAutolink += "=]"

        if not self.linkText:
            self.linkText = self.lt

        attrs = {
            "data-link-type":"dfn",
            "for":self.linkFor,
            "lt":self.lt,
            "bs-autolink-syntax":self.bsAutolink}
        return steps.Success(
            E.a(attrs, linkText))


DfnShorthand.startRe = re.compile(r"""
    (\\)?
    \[=
    (?!\s)(?:([^=|]*)/)?
    ([^\"=]+?)
    (?:\|)?""", re.X)

endRe = re.compile("=]")



# -*- coding: utf-8 -*-


from .BoolSet import BoolSet
from .dfnTypes import *
from .main import *
from .Nil import Nil
from .printjson import printjson
from .retrieve import DataFileRequester, defaultRequester, retrieveBoilerplateFile
from .status import *



# -*- coding: utf-8 -*-

import certifi
import io
import json
import re
import retrying
import os
from collections import defaultdict
from contextlib import closing
from json_home_client import Client as APIClient

from .. import config
from ..messages import *


def update(path, dryRun=False):
    say("Downloading anchor data...")
    shepherd = APIClient("https://api.csswg.org/shepherd/", version="vnd.csswg.shepherd.v1", ca_cert_path=certifi.where())
    rawSpecData = dataFromApi(shepherd, "specifications", draft=True)
    if not rawSpecData:
        return

    specs = dict()
    anchors = defaultdict(list)
    headings = defaultdict(dict)
    lastMsgTime = 0
    for i,rawSpec in enumerate(rawSpecData.values(), 1):
        lastMsgTime = config.doEvery(s=5, lastTime=lastMsgTime,
            action=lambda:say(f"Downloading data for spec {i}/{len(rawSpecData)}..."))
        rawSpec = dataFromApi(shepherd, 'specifications', draft=True, anchors=True, spec=rawSpec['name'])
        spec = genSpec(rawSpec)
        specs[spec['vshortname']] = spec
        specHeadings = headings[spec['vshortname']]

        def setStatus(obj, status):
            obj['status'] = status
            return obj
        rawAnchorData = (
            [setStatus(x, "snapshot") for x in linearizeAnchorTree(rawSpec.get('anchors', []))] +
            [setStatus(x, "current") for x in linearizeAnchorTree(rawSpec.get('draft_anchors', []))])
        for rawAnchor in rawAnchorData:
            rawAnchor = fixupAnchor(rawAnchor)
            linkingTexts = rawAnchor['linking_text']
            if linkingTexts[0] is None:
                # Happens if it had no linking text at all originally
                continue
            if len(linkingTexts) == 1 and linkingTexts[0].strip() == "":
                # Happens if it was marked with an empty lt and Shepherd still picked it up
                continue
            if 'section' in rawAnchor and rawAnchor['section'] == True:
                addToHeadings(rawAnchor, specHeadings, spec=spec)
            if rawAnchor['type'] not in ["heading"]:
                addToAnchors(rawAnchor, anchors, spec=spec)

    cleanSpecHeadings(headings)

    methods = extractMethodData(anchors)
    fors = extractForsData(anchors)

    if not dryRun:
        writtenPaths = set()
        try:
            p = os.path.join(path, "specs.json")
            writtenPaths.add(p)
            with io.open(p, 'w', encoding="utf-8") as f:
                f.write(json.dumps(specs, ensure_ascii=False, indent=2, sort_keys=True))
        except Exception as e:
            die("Couldn't save spec database to disk.\n{0}", e)
            return
        try:
            for spec, specHeadings in headings.items():
                p = os.path.join(path, "headings", "headings-{0}.json".format(spec))
                writtenPaths.add(p)
                with io.open(p, 'w', encoding="utf-8") as f:
                    f.write(json.dumps(specHeadings, ensure_ascii=False, indent=2, sort_keys=True))
        except Exception as e:
            die("Couldn't save headings database to disk.\n{0}", e)
            return
        try:
            writtenPaths.update(writeAnchorsFile(anchors, path))
        except Exception as e:
            die("Couldn't save anchor database to disk.\n{0}", e)
            return
        try:
            p = os.path.join(path, "methods.json")
            writtenPaths.add(p)
            with io.open(p, 'w', encoding="utf-8") as f:
                f.write(json.dumps(methods, ensure_ascii=False, indent=2, sort_keys=True))
        except Exception as e:
            die("Couldn't save methods database to disk.\n{0}", e)
            return
        try:
            p = os.path.join(path, "fors.json")
            writtenPaths.add(p)
            with io.open(p, 'w', encoding="utf-8") as f:
                f.write(json.dumps(fors, ensure_ascii=False, indent=2, sort_keys=True))
        except Exception as e:
            die("Couldn't save fors database to disk.\n{0}", e)
            return

    say("Success!")
    return writtenPaths


@retrying.retry(
    stop_max_attempt_number=3,
    wait_fixed=1000,
    # don't catch Ctrl-D, etc
    retry_on_exception=lambda x:isinstance(x, Exception))
def dataFromApi(api, *args, **kwargs):
    anchorDataContentTypes = [
        "application/json",
        "application/vnd.csswg.shepherd.v1+json",
        ]
    res = api.get(*args, **kwargs)
    if not res:
        raise Exception("Unknown error fetching anchor data. This might be transient; try again in a few minutes, and if it's still broken, please report it on GitHub.")
    data = res.data
    if res.status_code == 406:
        raise Exception("This version of the anchor-data API is no longer supported. Try updating Bikeshed. If the error persists, please report it on GitHub.")
    if res.content_type not in anchorDataContentTypes:
        raise Exception("Unrecognized anchor-data content-type '{0}'.", res.contentType)
    if res.status_code >= 300:
        raise Exception(f"Unknown error fetching anchor data; got status {res.status_code} and bytes:\n{data.decode('utf-8')}")
    if isinstance(data, bytes):
        raise Exception(f"Didn't get expected JSON data. Got:\n{data.decode('utf-8')}")
    return data


def linearizeAnchorTree(multiTree, list=None):
    if list is None:
        list = []
    # Call with multiTree being a list of trees
    for item in multiTree:
        if item['type'] in config.dfnTypes.union(["dfn", "heading"]):
            list.append(item)
        if item.get('children'):
            linearizeAnchorTree(item['children'], list)
            del item['children']
    return list


def genSpec(rawSpec):
    spec = {
        'vshortname': rawSpec['name'],
        'shortname': rawSpec.get('short_name'),
        'snapshot_url': rawSpec.get('base_uri'),
        'current_url': rawSpec.get('draft_uri'),
        'title': rawSpec.get('title'),
        'description': rawSpec.get('description'),
        'work_status': rawSpec.get('work_status'),
        'working_group': rawSpec.get('working_group'),
        'domain': rawSpec.get('domain'),
        'status': rawSpec.get('status'),
        'abstract': rawSpec.get('abstract')
    }
    if spec['shortname'] is not None and spec['vshortname'].startswith(spec['shortname']):
        # S = "foo", V = "foo-3"
        # Strip the prefix
        level = spec['vshortname'][len(spec['shortname']):]
        if level.startswith("-"):
            level = level[1:]
        if level.isdigit():
            spec['level'] = int(level)
        else:
            spec['level'] = 1
    elif spec['shortname'] is None and re.match(r"(.*)-(\d+)", spec['vshortname']):
        # S = None, V = "foo-3"
        match = re.match(r"(.*)-(\d+)", spec['vshortname'])
        spec['shortname'] = match.group(1)
        spec['level'] = int(match.group(2))
    else:
        spec['shortname'] = spec['vshortname']
        spec['level'] = 1
    return spec


def fixupAnchor(anchor):
    '''Miscellaneous fixes to the anchors before I start processing'''

    # This one issue was annoying
    if anchor.get('title', None) == "'@import'":
        anchor['title'] = "@import"

    # css3-tables has this a bunch, for some strange reason
    if anchor.get('uri', "").startswith("??"):
        anchor['uri'] = anchor['uri'][2:]

    # If any smart quotes crept in, replace them with ASCII.
    linkingTexts = anchor.get('linking_text', [anchor.get('title')])
    for i,t in enumerate(linkingTexts):
        if t is None:
            continue
        if "’" in t or "‘" in t:
            t = re.sub(r"‘|’", "'", t)
            linkingTexts[i] = t
        if "“" in t or "”" in t:
            t = re.sub(r"“|”", '"', t)
            linkingTexts[i] = t
    anchor['linking_text'] = linkingTexts

    # Normalize whitespace to a single space
    for k,v in list(anchor.items()):
        if isinstance(v, str):
            anchor[k] = re.sub(r"\s+", " ", v.strip())
        elif isinstance(v, list):
            for k1, v1 in enumerate(v):
                if isinstance(v1, str):
                    anchor[k][k1] = re.sub(r"\s+", " ", v1.strip())
    return anchor


def addToHeadings(rawAnchor, specHeadings, spec):
    uri = rawAnchor['uri']
    if uri[0] == "#":
        # Either single-page spec, or link on the top page of a multi-page spec
        heading = {
            'url': spec["{0}_url".format(rawAnchor['status'])] + uri,
            'number': rawAnchor['name'] if re.match(r"[\d.]+$", rawAnchor['name']) else "",
            'text': rawAnchor['title'],
            'spec': spec['title']
        }
        fragment = uri
        shorthand = "/" + fragment
    else:
        # Multi-page spec, need to guard against colliding IDs
        if "#" in uri:
            # url to a heading in the page, like "foo.html#bar"
            match = re.match(r"([\w-]+).*?(#.*)", uri)
            if not match:
                die("Unexpected URI pattern '{0}' for spec '{1}'. Please report this to the Bikeshed maintainer.", uri, spec['vshortname'])
                return
            page, fragment = match.groups()
            page = "/" + page
        else:
            # url to a page itself, like "foo.html"
            page, _, _ = uri.partition(".")
            page = "/" + page
            fragment = "#"
        shorthand = page + fragment
        heading = {
            'url': spec["{0}_url".format(rawAnchor['status'])] + uri,
            'number': rawAnchor['name'] if re.match(r"[\d.]+$", rawAnchor['name']) else "",
            'text': rawAnchor['title'],
            'spec': spec['title']
        }
    if shorthand not in specHeadings:
        specHeadings[shorthand] = {}
    specHeadings[shorthand][rawAnchor['status']] = heading
    if fragment not in specHeadings:
        specHeadings[fragment] = []
    if shorthand not in specHeadings[fragment]:
        specHeadings[fragment].append(shorthand)


def cleanSpecHeadings(headings):
    '''Headings data was purposely verbose, assuming collisions even when there wasn't one.
       Want to keep the collision data for multi-page, so I can tell when you request a non-existent page,
       but need to collapse away the collision stuff for single-page.'''
    for specHeadings in headings.values():
        for k, v in list(specHeadings.items()):
            if k[0] == "#" and len(v) == 1 and v[0][0:2] == "/#":
                # No collision, and this is either a single-page spec or a non-colliding front-page link
                # Go ahead and collapse them.
                specHeadings[k] = specHeadings[v[0]]
                del specHeadings[v[0]]


def addToAnchors(rawAnchor, anchors, spec):
    anchor = {
        'status': rawAnchor['status'],
        'type': rawAnchor['type'],
        'spec': spec['vshortname'],
        'shortname': spec['shortname'],
        'level': int(spec['level']),
        'export': rawAnchor.get('export', False),
        'normative': rawAnchor.get('normative', False),
        'url': spec["{0}_url".format(rawAnchor['status'])] + rawAnchor['uri'],
        'for': rawAnchor.get('for', [])
    }
    for text in rawAnchor['linking_text']:
        if anchor['type'] in config.lowercaseTypes:
            text = text.lower()
        text = re.sub(r'\s+', ' ', text)
        anchors[text].append(anchor)


def extractMethodData(anchors):
    '''Compile a db of {argless methods => {argfull method => {args, fors, url, shortname}}'''

    methods = defaultdict(dict)
    for key, anchors_ in anchors.items():
        # Extract the name and arguments
        match = re.match(r"([^(]+)\((.*)\)", key)
        if not match:
            continue
        methodName, argstring = match.groups()
        arglessMethod = methodName + "()"
        args = [x.strip() for x in argstring.split(",")] if argstring else []
        for anchor in anchors_:
            if anchor['type'] not in config.idlMethodTypes:
                continue
            if key not in methods[arglessMethod]:
                methods[arglessMethod][key] = {"args":args, "for": set(), "shortname":anchor['shortname']}
            methods[arglessMethod][key]["for"].update(anchor["for"])
    # Translate the "for" set back to a list for JSONing
    for signatures in methods.values():
        for signature in signatures.values():
            signature["for"] = sorted(signature["for"])
    return methods


def extractForsData(anchors):
    '''Compile a db of {for value => dict terms that use that for value}'''

    fors = defaultdict(set)
    for key, anchors_ in anchors.items():
        for anchor in anchors_:
            for for_ in anchor["for"]:
                if for_ == "":
                    continue
                fors[for_].add(key)
            if not anchor["for"]:
                fors["/"].add(key)
    for key, val in list(fors.items()):
        fors[key] = sorted(val)
    return fors


def writeAnchorsFile(anchors, path):
    '''
    Keys may be duplicated.

    key
    type
    spec
    shortname
    level
    status
    url
    export (boolish string)
    normative (boolish string)
    for* (one per line, unknown #)
    - (by itself, ends the segment)
    '''
    writtenPaths = set()
    groupedEntries = defaultdict(dict)
    for key,entries in anchors.items():
        group = config.groupFromKey(key)
        groupedEntries[group][key] = entries
    for group, anchors in groupedEntries.items():
        p = os.path.join(path, "anchors", "anchors-{0}.data".format(group))
        writtenPaths.add(p)
        with io.open(p, 'w', encoding="utf-8") as fh:
            for key, entries in sorted(anchors.items(), key=lambda x:x[0]):
                for e in entries:
                    fh.write(key + "\n")
                    for field in ["type", "spec", "shortname", "level", "status", "url"]:
                        fh.write(str(e.get(field, "")) + "\n")
                    for field in ["export", "normative"]:
                        if e.get(field, False):
                            fh.write("1\n")
                        else:
                            fh.write("\n")
                    for forValue in e.get("for", []):
                        if forValue:  # skip empty strings
                            fh.write(forValue + "\n")
                    fh.write("-" + "\n")
    return writtenPaths



# -*- coding: utf-8 -*-

from ..html import *
from ..messages import *

def exampleIDs(doc):
    '''
    Checks that every example in the document has an ID.
    '''
    if not doc.md.complainAbout['missing-example-ids']:
        return
    for el in findAll(".example:not([id])", doc):
        warn("Example needs ID:\n{0}", outerHTML(el)[0:100], el=el)



# -*- coding: utf-8 -*-


import io
import os

from ..InputSource import InputSource
from ..messages import *
from .main import scriptPath


class DataFileRequester(object):
    def __init__(self, type=None, fallback=None):
        self.type = type
        if self.type not in ("readonly", "latest"):
            raise Exception("Bad value for DataFileRequester.type, got '{0}'.", type)
        # fallback is another requester, used if the main one fails.
        self.fallback = fallback

    def fetch(self, *segs, **kwargs):
        str = kwargs.get("str", False)
        okayToFail = kwargs.get("okayToFail", False)
        fileType = kwargs.get("type", self.type)
        location = self._buildPath(segs=segs, fileType=fileType)
        try:
            if str:
                with io.open(location, "r", encoding="utf-8") as fh:
                    return fh.read()
            else:
                return io.open(location, "r", encoding="utf-8")
        except IOError:
            if self.fallback:
                try:
                    return self.fallback.fetch(*segs, str=str, okayToFail=okayToFail)
                except IOError:
                    return self._fail(location, str, okayToFail)
            return self._fail(location, str, okayToFail)

    def walkFiles(self, *segs, **kwargs):
        fileType = kwargs.get("type", self.type)
        for root, dirs, files in os.walk(self._buildPath(segs, fileType=fileType)):
            for file in files:
                yield file

    def _buildPath(self, segs, fileType=None):
        if fileType is None:
            fileType = self.fileType
        if fileType == "readonly":
            return scriptPath("spec-data", "readonly", *segs)
        else:
            return scriptPath("spec-data", *segs)

    def _fail(self, location, str, okayToFail):
        if okayToFail:
            if str:
                return ""
            else:
                return io.StringIO("")
        raise IOError("Couldn't find file '{0}'".format(location))

defaultRequester = DataFileRequester(type="latest", fallback=DataFileRequester(type="readonly"))


def retrieveBoilerplateFile(doc, name, group=None, status=None, error=True):
    # Looks in three locations, in order:
    # the folder the spec source is in, the group's boilerplate folder, and the generic boilerplate folder.
    # In each location, it first looks for the file specialized on status, and then for the generic file.
    # Filenames must be of the format NAME.include or NAME-STATUS.include
    if group is None and doc.md.group is not None:
        group = doc.md.group.lower()
    if status is None and doc.md.rawStatus is not None:
        megaGroup,_,status = doc.md.rawStatus.partition("/")
        if status == "":
            status = megaGroup

    searchLocally = doc.md.localBoilerplate[name]

    def boilerplatePath(*segs):
        return scriptPath("boilerplate", *segs)
    statusFile = "{0}-{1}.include".format(name, status)
    genericFile = "{0}.include".format(name)
    sources = []
    if searchLocally:
        sources.append(doc.inputSource.relative(statusFile))  # Can be None.
        sources.append(doc.inputSource.relative(genericFile))
    else:
        for f in (statusFile, genericFile):
            if doc.inputSource.cheaplyExists(f):
                warn(("Found {0} next to the specification without a matching\n"+
                    "Local Boilerplate: {1} yes\n"+
                    "in the metadata. This include won't be found when building via a URL.").format(f, name))
                # We should remove this after giving specs time to react to the warning:
                sources.append(doc.inputSource.relative(f))
    if group:
        sources.append(InputSource(boilerplatePath(group, statusFile)))
        sources.append(InputSource(boilerplatePath(group, genericFile)))
    sources.append(InputSource(boilerplatePath(statusFile)))
    sources.append(InputSource(boilerplatePath(genericFile)))

    # Watch all the possible sources, not just the one that got used, because if
    # an earlier one appears, we want to rebuild.
    doc.recordDependencies(*sources)

    for source in sources:
        if source is not None:
            try:
                return source.read().content
            except IOError:
                # That input doesn't exist.
                pass
    else:
        if error:
            die("Couldn't find an appropriate include file for the {0} inclusion, given group='{1}' and status='{2}'.", name, group, status)
        return ""



# -*- coding: utf-8 -*-


from .main import fixupDataFiles, update, updateReadonlyDataFiles
from .manifest import createManifest

__all__ = ["updateBackRefs", "updateCrossRefs", "updateBiblio", "updateCanIUse", "updateLinkDefaults", "updateTestSuites", "updateLanguages"]



# -*- coding: utf-8 -*-

import io
import json
import os
import requests
from collections import OrderedDict

from ..messages import *


def update(path, dryRun=False):
    say("Downloading Can I Use data...")
    try:
        response = requests.get("https://raw.githubusercontent.com/Fyrd/caniuse/master/fulldata-json/data-2.0.json")
    except Exception as e:
        die("Couldn't download the Can I Use data.\n{0}", e)
        return

    try:
        data = response.json(encoding="utf-8", object_pairs_hook=OrderedDict)
    except Exception as e:
        die("The Can I Use data wasn't valid JSON for some reason. Try downloading again?\n{0}", e)
        return

    basicData = {"agents":[], "features":{}, "updated": data["updated"]}

    # Trim agent data to minimum required - mapping codename to full name
    codeNames = {}
    agentData = {}
    for codename,agent in data["agents"].items():
        codeNames[codename] = agent["browser"]
        agentData[agent["browser"]] = codename
    basicData["agents"] = agentData

    # Trim feature data to minimum - notes and minimum supported version
    def simplifyStatus(s, *rest):
        if "x" in s or "d" in s or "n" in s or "p" in s:
            return "n"
        elif "a" in s:
            return "a"
        elif "y" in s:
            return "y"
        elif "u" in s:
            return "u"
        else:
            die("Unknown CanIUse Status '{0}' for {1}/{2}/{3}. Please report this as a Bikeshed issue.", s, *rest)
            return None
    def simplifyVersion(v):
        if "-" in v:
            # Use the earliest version in a range.
            v,_,_ = v.partition("-")
        return v
    featureData = {}
    for featureName,feature in data["data"].items():
        notes = feature["notes"]
        url = feature["spec"]
        basicData["features"][featureName] = url
        browserData = {}
        for browser,versions in feature["stats"].items():
            descendingVersions = list(reversed(versions.items()))
            mostRecent = descendingVersions[0]
            version = simplifyVersion(mostRecent[0])
            status = simplifyStatus(mostRecent[1], featureName, browser, version)
            if status == "n":
                # Most recent version is broken, so we're done
                pass
            elif status == "u":
                # Seek backwards until I find something other than "u"
                for v,s in descendingVersions:
                    if simplifyStatus(s) != "u":
                        status = simplifyStatus(s)
                        version = simplifyVersion(v)
                        break
            else:
                # Status is either (a)lmost or (y)es,
                # seek backwards thru time as long as it's the same.
                for v,s in descendingVersions:
                    if simplifyStatus(s) == status:
                        version = simplifyVersion(v)
                    else:
                        break
            browserData[codeNames[browser]] = "{0} {1}".format(status, version)
        featureData[featureName] = {"notes":notes, "url":url, "support":browserData}

    writtenPaths = set()
    if not dryRun:
        try:
            p = os.path.join(path, "caniuse", "data.json")
            writtenPaths.add(p)
            with io.open(p, 'w', encoding="utf-8") as fh:
                fh.write(json.dumps(basicData, indent=1, ensure_ascii=False, sort_keys=True))

            for featureName, feature in featureData.items():
                p = os.path.join(path, "caniuse", "feature-{0}.json".format(featureName))
                writtenPaths.add(p)
                with io.open(p, 'w', encoding='utf-8') as fh:
                    fh.write(json.dumps(feature, indent=1, ensure_ascii=False, sort_keys=True))
        except Exception as e:
            die("Couldn't save Can I Use database to disk.\n{0}", e)
            return
    say("Success!")
    return writtenPaths



# -*- coding: utf-8 -*-



class Repository(object):
    '''
    A class for representing spec repositories.
    '''

    def __init__(self, url, name=None, type=None):
        self.ns = "com"
        self.url = url
        self.api = "https://api.github.com"
        if name:
            self.name = name
        else:
            self.name = url

        if type:
            self.type = type
        else:
            self.type = "unknown"

    def formatIssueUrl(self, *args, **kwargs):
        # Dunno how to format an arbitrary issue url,
        # so give up and just point to the repo.
        return self.url


class GithubRepository(Repository):
    def __init__(self, ns, user, repo):
        super(GithubRepository, self).__init__("https://github.{0}/{1}/{2}".format(ns, user, repo), "{0}/{1}".format(user, repo))
        self.ns = ns
        self.user = user
        self.repo = repo
        self.type = "github"
        if ns == "com":
            self.api = "https://api.github.com"
        else:
            self.api = "https://github.{0}/api/v3".format(ns)

    def formatIssueUrl(self, id=None):
        if id is None:
            return "https://github.{0}/{1}/{2}/issues/".format(self.ns, self.user, self.repo)
        return "https://github.{0}/{1}/{2}/issues/{3}".format(self.ns, self.user, self.repo, id)

    def __str__(self):
        return "{0}/{1}".format(self.user, self.repo)



# -*- coding: utf-8 -*-

class StringEnum(object):
	'''
	Simple Enum class.

	Takes strings, returns an object that acts set-like (you can say `"foo" in someEnum`),
	but also hangs the strings off of the object as attributes (you can say `someEnum.foo`)
	or as a dictionary (`someEnum["foo"]`).
	'''

	def __init__(self, *vals):
		self._vals = {}
		for val in vals:
			self._vals[val] = val
			setattr(self, val, val)

	def __iter__(self):
		return iter(self._vals)

	def __len__(self):
		return len(self._vals)

	def __contains__(self, val):
		return val in self._vals

	def __getitem__(self, val):
		return self._vals[val]


# -*- coding: utf-8 -*-


import copy

import attr

from .utils import stripLineBreaks

@attr.s(slots=True)
class RefWrapper(object):
    # Refs don't contain their own name, so I don't have to copy as much when there are multiple linkTexts
    # This wraps that, producing an object that looks like it has a text property.
    # It also makes all the ref dict keys look like object attributes.

    text = attr.ib()
    _ref = attr.ib()

    @property
    def type(self):
        return decode(self._ref['type'].strip())
    @property
    def spec(self):
        return decode(self._ref['spec'].strip())
    @property
    def shortname(self):
        return decode(self._ref['shortname'].strip())
    @property
    def level(self):
        if self._ref['level'] is None:
            return ""
        return decode(self._ref['level'].strip())
    @property
    def status(self):
        return decode(self._ref['status'].strip())
    @property
    def url(self):
        return decode(self._ref['url'].strip())
    @property
    def export(self):
        return self._ref['export']
    @property
    def normative(self):
        return self._ref['normative']
    @property
    def for_(self):
        return [x.strip() for x in self._ref['for']]
    @property
    def el(self):
        return self._ref.get('el', None)

    '''
        "type": linesIter.next(),
        "spec": linesIter.next(),
        "shortname": linesIter.next(),
        "level": linesIter.next(),
        "status": linesIter.next(),
        "url": linesIter.next(),
        "export": linesIter.next() != "\n",
        "normative": linesIter.next() != "\n",
        "for": [],
        (optionall) "el": manuallyProvided,
    '''

    def __json__(self):
        refCopy = copy.copy(self._ref)
        refCopy['text'] = self.text
        return stripLineBreaks(refCopy)

def decode(s):
    # TODO: verify that this can be removed
    return s
